function [Sigma, s, perf,niters,gaps] = compute_sigma_fdiv(M,N,epsilon,n_init,maxiter)

% n_init = number of initialization of r (at least 4)
% maxiter = number of steps in concave convex procedure (4 is enough)

d =  size(M,1);

% starts with a uniform grid on [0,1]
ss_grid = ( 0:(n_init-1) )  / ( n_init - 1);
for igrid = 1:length(ss_grid)    
    if igrid==1
        % first_itearation: no warm start
        [Z,Y,gap,niter] = maximize_sos_chambolle_pock(ss_grid(igrid)*M+(1-ss_grid(igrid))*N,epsilon,2000);
    elseif igrid==2
        % then: warm start with no prediction correction
        [Z,Y,gap,niter] = maximize_sos_chambolle_pock(ss_grid(igrid)*M+(1-ss_grid(igrid))*N,epsilon,2000,Y,Z);
    else
        % then: warm start with prediction correction        
        [Z,Y,gap,niter] = maximize_sos_chambolle_pock(ss_grid(igrid)*M+(1-ss_grid(igrid))*N,epsilon,2000,2*Y-reshape(Ys(:,igrid-2),[d d]),2*Z-reshape(Zs(:,igrid-2),[d d]));
    end
    
    niters(igrid)=niter;
    gaps(igrid)=gap;
    
    % compute cost fdunctions
    Zs(:,igrid) = Z(:);
    Ys(:,igrid) = Y(:);
    rr = ss_grid(igrid) / ( 1-ss_grid(igrid));
    if rr == 0
        perfs(igrid) = ( trace(Z*N) - 1) / (rr+1);
    elseif rr == Inf
        perfs(igrid) = -Inf;
    else
        perfs(igrid) = ( trace(Z*(rr*M+N )) - rr*log(rr) + rr - 1) / (rr+1);
    end
    % if mod(igrid,100)==0, igrid, end
end


% find best perf
[a,igrid] = max(perfs);
Z = reshape(Zs(:,igrid),d,d);
Y = reshape(Zs(:,igrid),d,d);
ss = ss_grid(igrid);
rr = ss/(1-ss);

% launch iterative improvement algorithm
for iter = 1:maxiter
    % compute s
    [Z,Y,gap,niter] = maximize_sos_chambolle_pock(ss*M+(1-ss)*N,epsilon,2000,Y,Z);

    % update r
    u = trace(Z*M)-trace(Z*N);
    if u>0
        rr = 1;
        for jj=1:6  % 6 iterations enough
            rr = (2 - log(rr) + u )./ ( 1 + 1./rr);
        end
    else
        rr = 1;
        for jj=1:6   % 6 iterations enough
            rr  =exp( ( 1+u + rr .* ( log(rr)-1))./(1+rr) );
        end
    end
    ss = rr / (rr+1);
    
end

Sigma = Z;
s = ss;
rr = ss/(1-ss);
if rr == 0
    perf = ( trace(Z*N) - 1) / (rr+1);
elseif rr == Inf
    perf = -Inf;
else
    perf = ( trace(Z*(rr*M+N )) - rr*log(rr) + rr - 1) / (rr+1);
end
