% plot figure 1 (left)

clear all

imodel = 1

seed=1;
randn('state',seed);
rand('state',seed);

switch imodel
    case 1,
        model = 'trigo_1d';
    case 2,
        model = 'trigo_multipled';
    case 3,
        model = 'discrete';
end

% parameters of the model
switch model
    case 'trigo_1d'
        
        x  = (-100000:100000)/100000; % used to compute true value
        px = 2/pi * sqrt(1-x.*x);
        qx = 1/2;
        Dpq_value = mean( qx.* ( (px./qx ) .* log(px./qx + 1e-14 ) - px./qx  + 1 ))*2
        rrrs = 0:16;
        
        
end



SOS = [];
SOS_kelley = [];
QT = [];
QT_diag= [];
QT_old = [];

for irr=1:length(rrrs)
    
    r=rrrs(irr);
    
    if r==0
        
        QT(irr) = 0;
        QT_diag(irr) = 0;
        QT_old(irr) = 0;
        SOS(irr) = 0;
        SOS_kelley(irr) = 0;
        
    else
        
        
        switch model
            case 'trigo_1d'
                
                
                d = 2*r+1;
                
                %% compute moments
                B = eye(2*r+1);  % moments of the uniform distribution (q)
                A = ones(2*r+1); % moments of the first distrution (p)
                for i=1:(2*r+1)
                    for j=1:(2*r+1)
                        if (i~=j)
                            A(i,j) = 2 * besselj(1,(i-j)*pi) / (i-j) / pi * (-1)^(i-j);
                        end
                    end
                end
                A = A / (2*r+1);
                B = B / (2*r+1);
                
                
                
                %% create Toeplitz projection matrix as global variable
                m = 2*r+1;
                global TP
                TP = zeros(m*m,m*m);
                for i=1:m, for j=1:m
                        AA = zeros(m,m);
                        AA(i,j)=1;
                        BB = toeplitz_projection(AA);
                        TP(:,(i-1)*m+j) = BB(:);
                    end
                end
                TP = sparse(TP);
                
                %% create generator of Toeplitz matrices
                global TPg
                TPg = zeros(m*m,2*r+1);
                
                for i=1:2*r+1
                    q = zeros(2*r+1,1); q(i)=1;
                    q = toeplitz(q);
                    TPg(:,i) = q(:);
                end
                TPg = sparse(TPg);
                
        end
        
        
        
        
        
        
        %% quantum information
        MM = inv(sqrtm(B))*A*inv(sqrtm(B));
        AA = sqrtm(B) * ( MM * logm(MM) - MM + eye(d) ) * sqrtm(B);
        
        
        %% optimizing quantum
        cvx_begin
        variable mus(size(TPg,2),1)
        variable SSS(d,d)
        dual variable V;
        minimize ( trace(SSS) )
        subject to
        for i=1:d
            for j=1:i
                SSS(i,j) == TPg( (i-1)*d+j,:) * mus;
            end
        end
        V: SSS - AA == semidefinite(d);
        cvx_end
        
        
        V = (V+V')/2;
        QT(irr) =  trace(SSS)
        
        %% optimizing quantum-diag
        [a,b] = max(diag(AA));
        QT_diag(irr) = a*d
        
        %% old Bach 2022 paper
        eta = ones(d,1)*d;
        Asqrtm = sqrtm(A);
        for iter=1:1000
            eta = exp( d * diag( Asqrtm * logm ( Asqrtm * diag(eta) * Asqrtm + 1e-12 * eye(d) ) * Asqrtm ) );
            eta = real(d * eta / sum(eta));
        end
        
        Anew =  diag(sqrt(eta)) * A *  diag(sqrt(eta));
        QT_old(irr) = trace ( Anew * logm( Anew + 1e-12 * eye(d) ) - Anew * diag(log(eta/d)) - Anew + diag(eta/d) )
        
        
        %% initialize matrix M from quantum
        V = V + ( max(-min(eig(V)),0) + 1e-12 ) * eye(d);
        % find the optimal M and M for the quantum case
        T = chol(V);
        At = T*A*T';
        Bt = T*B*T';
        [u,s,v] = svd(inv(sqrtm(Bt))*sqrtm(At));
        s =  (diag(s)).^2;
        M = 0;
        N = 0;
        for i=1:d
            for j=1:d
                if abs(s(i)-s(j))>1e-10,
                    M = M + ( s(i) * log( s(i) ) - s(i) - s(j) * log( s(j) ) + s(j) )/(s(i)-s(j)) * u(:,i)'*Bt*u(:,j) * u(:,i)*u(:,j)';
                    N = N + (log(s(i)) +1/s(i) - log(s(j))- 1/s(j) )/( 1/s(i) - 1/s(j) ) * v(:,i)'*At*v(:,j) * v(:,i)*v(:,j)';
                else
                    M = M + log(s(i)) * u(:,i)'*Bt*u(:,j) * u(:,i)*u(:,j)';
                    N = N + (1 - s(i) )  * v(:,i)'*At*v(:,j) * v(:,i)*v(:,j)';
                end
            end
        end
        M = T' * inv(sqrtm(Bt)) * M * inv(sqrtm(Bt)) * T;
        N = T' * inv(sqrtm(At)) * N * inv(sqrtm(At)) * T;
        M = (M + M') /2;
        N = (N + N') /2;
        
        %% parameters to compute function
        n_init = 2000;
        maxiter = 200;
        epsilon = 1e-3;
        
        %% SOS without optimization
        
        [Sigma, s, val,niters, gaps] = compute_sigma_fdiv(M,N,epsilon,n_init,maxiter);
        SOS(irr) =    trace(A*M)+trace(B*N) - 2 * val;
        [ SOS', QT', QT_diag', QT_old']
        
        %% kelley's method
        value = [];
        value_upper = [];
        current_gaps = [];
        m = size(M,1);
        M0 = M; N0 = N;
        grad_Ms = {};
        grad_Ns = {};
        Delta = ( trace(A*M)+trace(B*N) ) * 2;
        maxiter_subg = 256;
        for iter_subg = 1:maxiter_subg
            Ms{iter_subg} = M;
            Ns{iter_subg} = N;
            
            [Sigma, s, val, gaps] = compute_sigma_fdiv(M,N,epsilon,n_init,maxiter);
            value(iter_subg) =    trace(A*M)+trace(B*N) - 2 * val;
            grad_Ms{iter_subg} = A - 2*s*Sigma;
            grad_Ns{iter_subg} = B - 2*(1-s)*Sigma;
            
            
            
            mloc = size(TPg,2);
            
            cvx_begin quiet
            variable DeltaM(mloc,1)
            variable DeltaN(mloc,1)
            variable t(1,1);
            maximize ( t )
            subject to
            for i=1:iter_subg
                value(i) + trace(grad_Ms{i}*(M0+reshape(TPg*DeltaM,[m m])-Ms{i})) + trace(grad_Ns{i}*(N0+reshape(TPg*DeltaN,[m m])-Ns{i})) >= t;
            end
            norm(reshape(TPg*DeltaM,[m m])) + norm(reshape(TPg*DeltaN,[m m])) <= Delta
            cvx_end
            DeltaM = reshape(TPg*DeltaM,[m m]);
            DeltaN = reshape(TPg*DeltaN,[m m]);
            
            
            if mod(iter_subg,m+2)==0
                if norm(DeltaM) + norm(DeltaN) >= Delta * .999, 'increasing Delta', Delta = Delta *2;
                end
            end
            M = M0+ DeltaM ;
            N = N0+ DeltaN ;
            value_upper(iter_subg) = t;
            if ( norm( DeltaM ) + norm( DeltaN ) < Delta * .999)
                current_gaps(iter_subg) = t - max(value);
            else
                current_gaps(iter_subg) = Inf;
            end
            if ( norm( DeltaM ) + norm( DeltaN )  < Delta * .999) & ( ( t - max(value) ) < 1e-4 ), 'exit', break; end
        end
        
        
        SOS_kelley(irr) = max(value)
        
    end
    
    save(sprintf('data_entropy_kelley_small_%s',model))
    
end




%% plot
rrs = rrrs;
try
    openfig('saved_figure.fig');
catch
    'figure file not found'
end

plot(rrs,SOS*0+Dpq_value,'g','linewidth',2);
hold on

plot(rrs,SOS_kelley,'rx-','linewidth',2,'markersize',8);
plot(rrs,QT,'bx-','linewidth',2,'markersize',8);
plot(rrs,QT_diag,'bx:','linewidth',2,'markersize',8);
plot(rrs,QT_old,'kx-','linewidth',2,'markersize',8);
if imodel==1
    legend('D(p||q)','SOS/OPT','QT','QT-diag','QT-diag-[4]','location','southeast');
else
    legend('D(p||q)','SOS','QT','QT-diag','QT-diag-[4]','location','southeast');
    
end
hold off
set(gca,'fontsize',18)
ylabel('relative entropies');
xlabel('number of frequencies (r)');
axis([0 max(rrs) 0 Dpq_value*1.1])
print('-depsc', sprintf('figure_entropy_kelley_small_%s.eps',model));



