%% plots Figure 1 (right), Figure 2 (right), Figure 3 (right)

clear all

for imodel = 1:3
    
    seed=1;
    randn('state',seed);
    rand('state',seed);
    
    switch imodel
        case 1,
            model = 'trigo_1d';
        case 2,
            model = 'trigo_multipled';
        case 3,
            model = 'discrete';
    end
    
    % parameters of the model
    switch model
        case 'trigo_1d'
            
            x  = (-100000:100000)/100000; % used to compute true value
            px = 2/pi * sqrt(1-x.*x);
            qx = 1/2;
            Dpq_value = mean( qx.* ( (px./qx ) .* log(px./qx + 1e-14 ) - px./qx  + 1 ))*2
            rrrs = 0:48;
            
        case 'trigo_multipled'
            
            n = 3; % number of dimensions
            rhopp = .3; % parameter of the distribution
            
            MMM = zeros(n,n-1);
            for i=1:n-1
                MMM(i+1,1:i) = 1;
            end
            Dpq_value = (n-1) * log(1/rhopp)
            rrrs = 0:4;
            
        case 'discrete'
            
            n = 20;
            rhopp = 0.5;
            Dpq_value = (n-1) * ( ( 1- rhopp/2) * log(2-rhopp) + rhopp/2 * log(rhopp) )
            
            MMM = zeros(n,n-1);
            for i=1:n-1
                MMM(i+1,1:i) = 1;
            end
            
            rrrs = 0:2;
            
    end
    
    
    
    SOS = [];
    QT = [];
    QT_diag= [];
    QT_old = [];
    
    for irr=1:length(rrrs)
        
        r=rrrs(irr);
        
        if r==0
            
            QT(irr) = 0;
            QT_diag(irr) = 0;
            QT_old(irr) = 0;
            SOS(irr) = 0;
            
        else
            
            
            switch model
                case 'trigo_1d'
                    
                    
                    d = 2*r+1;
                    
                    %% compute moments
                    B = eye(2*r+1);  % moments of the uniform distribution (q)
                    A = ones(2*r+1); % moments of the first distrution (p)
                    for i=1:(2*r+1)
                        for j=1:(2*r+1)
                            if (i~=j)
                                A(i,j) = 2 * besselj(1,(i-j)*pi) / (i-j) / pi * (-1)^(i-j);
                            end
                        end
                    end
                    A = A / (2*r+1);
                    B = B / (2*r+1);
                    %% create Toeplitz projection matrix as global variable
                    m = 2*r+1;
                    global TP
                    TP = zeros(m*m,m*m);
                    for i=1:m, for j=1:m
                            AA = zeros(m,m);
                            AA(i,j)=1;
                            BB = toeplitz_projection(AA);
                            TP(:,(i-1)*m+j) = BB(:);
                        end
                    end
                    TP = sparse(TP);
                    
                    %% create generator of Toeplitz matrices
                    global TPg
                    TPg = zeros(m*m,2*r+1);
                    
                    for i=1:2*r+1
                        q = zeros(2*r+1,1); q(i)=1;
                        q = toeplitz(q);
                        TPg(:,i) = q(:);
                    end
                    TPg = sparse(TPg);
                    
                case 'trigo_multipled'
                    
                    
                    omegas = ind2subv(ones(1,n)*(2*r+1),1:(2*r+1)^n) - r-1;
                    d = (2*r+1)^n;
                    allomegadiffs = [];
                    for i=1:n
                        temp = omegas(:,i) * ones(1,d) - ones(d,1) * omegas(:,i)';
                        allomegadiffs = [ allomegadiffs, temp(:) ];
                    end
                    [allomegadiffsuniques,IA,IC] =  unique(allomegadiffs,'rows');
                    
                    
                    
                    %% compute moments
                    B = eye(d);  % moments of the uniform distribution (q)
                    A = ones(d); % moments of the first distrution (p)
                    for i=1:d
                        for j=1:d
                            if (i~=j)
                                if abs(sum(omegas(i,:)-omegas(j,:)))>1e-12,
                                    A(i,j) = 0;
                                else
                                    for k=1:n-1
                                        if ( MMM(:,k)'*omegas(i,:)' -  MMM(:,k)'*omegas(j,:)' ) ~= 0
                                            A(i,j) = A(i,j) * sin( ( MMM(:,k)'*omegas(i,:)' -  MMM(:,k)'*omegas(j,:)' ) * pi * rhopp ) / ( (  MMM(:,k)'*omegas(i,:)' -  MMM(:,k)'*omegas(j,:)' ) * pi * rhopp );
                                        end
                                        
                                    end
                                end
                            end
                            if 0
                                %% check moments
                                mm = 100000;
                                x = rand(1,mm);
                                y = x + (rand(1,mm)*2-1) * rhopp/2;
                                AA(i,j) = mean( exp( 2* sqrt(-1) * pi * ( omegas(i,:) -omegas(j,:) ) * [ x;y ]));
                                % if abs(A(i,j)-real(AA(i,j)))>1e-1, pause; end
                            end
                        end
                    end
                    B = B / (2*r+1)^n;
                    A = A / (2*r+1)^n;
                    
                    
                    %% create Toeplitz projection matrix as global variable        ;
                    global TP
                    TP = sparse(d*d,d*d);
                    for i = 1:size(allomegadiffsuniques,1)
                        ind = find( IC == i );
                        TP(ind,ind) = 1/length(ind);
                    end
                    
                    
                    %% create generator of Toeplitz matrices
                    global TPg
                    TPg = sparse(d*d,size(allomegadiffsuniques,1));
                    
                    for i=1:size(allomegadiffsuniques,1)
                        q = zeros(d,d); q(find( IC == i ))=1;
                        TPg(:,i) = q(:);
                    end
                    
                    
                    
                    
                case 'discrete'
                    
                    
                    omegas = zeros(1,n);
                    omegaloc = zeros(1,n);
                    for i = 1:r
                        omegalocnew = [];
                        for j=1:size(omegaloc,1)
                            ind = find( omegaloc(j,:)==0 );
                            for a=1:length(ind);
                                temp = omegaloc(j,:); temp(ind(a)) = 1;
                                omegalocnew = [omegalocnew ; temp ];
                            end
                        end
                        omegalocnew = unique(omegalocnew,'rows');
                        omegas = [ omegas; omegalocnew ];
                        omegaloc = omegalocnew;
                    end
                    d = size(omegas,1);
                    
                    
                    allomegadiffs = [];
                    for i=1:n
                        temp = abs(omegas(:,i) * ones(1,d) - ones(d,1) * omegas(:,i)');
                        allomegadiffs = [ allomegadiffs, temp(:) ];
                    end
                    [allomegadiffsuniques,IA,IC] =  unique(allomegadiffs,'rows');
                    
                    
                    %% compute moments
                    B = eye(d); % moments of the uniform distribution (q)
                    A = ones(d); % moments of the first distrution (p)
                    for i=1:d
                        for j=1:d
                            if (i~=j)
                                delta = abs(omegas(i,:)-omegas(j,:));
                                if mod(sum(delta),2)
                                    A(i,j) = 0;
                                else
                                    for k=1:n-1
                                        if mod(MMM(:,k)'*delta',2),  A(i,j) = A(i,j) *  (1-rhopp); end
                                    end
                                end
                            end
                        end
                    end
                    
                    A = A / d;
                    B = B / d;
                    
                    
                    %% create Toeplitz projection matrix as global variable        ;
                    global TP
                    TP = sparse(d*d,d*d);
                    for i = 1:size(allomegadiffsuniques,1)
                        ind = find( IC == i );
                        TP(ind,ind) = 1/length(ind);
                    end
                    
                    
                    %% create generator of Toeplitz matrices
                    global TPg
                    TPg = sparse(d*d,size(allomegadiffsuniques,1));
                    
                    for i=1:size(allomegadiffsuniques,1)
                        q = zeros(d,d); q(find( IC == i ))=1;
                        TPg(:,i) = q(:);
                    end
            end
            
            
            
            %% quantum information
            MM = inv(sqrtm(B))*A*inv(sqrtm(B));
            AA = sqrtm(B) * ( MM * logm(MM) - MM + eye(d) ) * sqrtm(B);
            
            
            %% optimizing quantum
            cvx_begin
            variable mus(size(TPg,2),1)
            variable SSS(d,d)
            dual variable V;
            minimize ( trace(SSS) )
            subject to
            for i=1:d
                for j=1:i
                    SSS(i,j) == TPg( (i-1)*d+j,:) * mus;
                end
            end
            V: SSS - AA == semidefinite(d);
            cvx_end
            
            
            V = (V+V')/2;
            QT(irr) =  trace(SSS)
            
            %% optimizing quantum-diag
            [a,b] = max(diag(AA));
            QT_diag(irr) = a*d
            
            %% old Bach 2022 paper
            eta = ones(d,1)*d;
            Asqrtm = sqrtm(A);
            for iter=1:1000
                eta = exp( d * diag( Asqrtm * logm ( Asqrtm * diag(eta) * Asqrtm + 1e-12 * eye(d) ) * Asqrtm ) );
                eta = real(d * eta / sum(eta));
            end
            
            Anew =  diag(sqrt(eta)) * A *  diag(sqrt(eta));
            QT_old(irr) = trace ( Anew * logm( Anew + 1e-12 * eye(d) ) - Anew * diag(log(eta/d)) - Anew + diag(eta/d) )
            
            
            
            
        end
        
        save(sprintf('data_entropy_large_%s',model))
        
    end
    
    
    
    %% plot
    rrs = rrrs;
    
    try
        openfig('saved_figure.fig');
    catch
        'figure file not found'
    end
    
    if imodel ==1
        % plot(rrs,SOS,'rx-','linewidth',2,'markersize',8);
        plot(log10(rrs),log10(Dpq_value-QT),'bx-','linewidth',2,'markersize',8);
        hold on
        plot(log10(rrs),log10(Dpq_value-QT_diag),'bx:','linewidth',2,'markersize',8);
        plot(log10(rrs),log10(Dpq_value-QT_old),'kx-','linewidth',2,'markersize',8);
        % legend('D(p||q)','SOS','QT','QT-diag','QT-diag-[4]','location','southeast');
        legend('QT','QT-diag','QT-diag-[4]','location','northeast');
        hold off
        set(gca,'fontsize',18)
        ylabel('log_{10}(D(p||q)-relative entropies)');
        xlabel('log_{10}(maximal frequency) (log_{10}(r))');
    else
        plot(rrs,QT*0+Dpq_value,'g','linewidth',2);
        hold on
        
        % plot(rrs,SOS,'rx-','linewidth',2,'markersize',8);
        plot(rrs,QT,'bx-','linewidth',2,'markersize',8);
        plot(rrs,QT_diag,'bx:','linewidth',2,'markersize',8);
        plot(rrs,QT_old,'kx-','linewidth',2,'markersize',8);
        % legend('D(p||q)','SOS','QT','QT-diag','QT-diag-[4]','location','southeast');
        if imodel == 2
            legend('D(p||q)','QT','QT-diag','QT-diag-[4]','location','southeast');
        else
            legend('D(p||q)','QT','QT-diag','QT-diag-[4]','location','northwest');            
        end
        hold off
        set(gca,'fontsize',18)
        ylabel('relative entropies');
        if imodel == 3
            xlabel('maximal cardinality (r)');
        else
            xlabel('maximal frequency (r)');
        end
        axis([0 max(rrs) 0 Dpq_value*1.1])
    end
    print('-depsc', sprintf('figure_entropy_large_%s.eps',model));
end


