% plot figure 4

clear all

for imodel = 1
    
    seed=1;
    randn('state',seed);
    rand('state',seed);
    
    switch imodel
        case 1,
            model = 'trigo_1d';
        case 2,
            model = 'trigo_multipled';
        case 3,
            model = 'discrete';
    end
    
    % parameters of the model
    switch model
        case 'trigo_1d'
            
            logparval = log(besseli(0,1))
            rrrs = 0:8;
    end
    
    QT_no_diag_recentered = [];
    QT_old_no_diag = [];
    QT_kelley_notV = [];
    QT_kelley_V = [];
    
    for irr=1:length(rrrs)
        
        r=rrrs(irr);
        
        if r==0
            QT_no_diag_recentered(irr) = 1;
            QT_old_no_diag(irr) = 1;
            QT_kelley_notV(irr) = 1;
            QT_kelley_V(irr) = 1;
        else
            
            
            switch model
                case 'trigo_1d'
                    
                    
                    d = 2*r+1;
                    
                    %% compute moments
                    B = eye(2*r+1);  % moments of the uniform distribution (q)
                    B = B / (2*r+1);
                    
                    H = zeros(2*r+1,2*r+1);
                    H(r,r+2)=1/2; H(r+2,r)=1/2;
                    H = H * (2*r+1);
                    
                    
                    
                    %% create Toeplitz projection matrix as global variable
                    m = 2*r+1;
                    global TP
                    TP = zeros(m*m,m*m);
                    for i=1:m, for j=1:m
                            AA = zeros(m,m);
                            AA(i,j)=1;
                            BB = toeplitz_projection(AA);
                            TP(:,(i-1)*m+j) = BB(:);
                        end
                    end
                    TP = sparse(TP);
                    
                    %% create generator of Toeplitz matrices
                    global TPg
                    TPg = zeros(m*m,2*r+1);
                    
                    for i=1:2*r+1
                        q = zeros(2*r+1,1); q(i)=1;
                        q = toeplitz(q);
                        TPg(:,i) = q(:);
                    end
                    TPg = sparse(TPg);
                    
            end
            
            
            
            %% QT_old_no_diag
            m = size(H,1);
            temp = expm( H + logm(B));
            A = temp / trace(temp);
            A = (A +A')/2;
            QT_old_no_diag(irr) = log(trace(temp));
            
            
            %% QT_no_diag_recentered
            H = reshape(TP*H(:),[m m]);
            
            temp = expm( H + logm(B));
            A = temp / trace(temp);
            A = (A +A')/2;
            QT_no_diag_recentered(irr) = log(trace(temp));
            
            % initializing for other optimizations
            [u,lambda] = eig(B);
            lambda = diag(lambda);
            N = -eye(m);
            for i=1:m
                for j=1:m
                    if (i==j) |  ( abs( lambda(i)-lambda(j))<1e-10 ) ,
                        N = N + u(:,i)' * A * u(:,j) / lambda(i) * u(:,i)*u(:,j)';
                    else
                        N = N + u(:,i)' * A * u(:,j) * ( log(lambda(i))-log(lambda(j))) / ( lambda(i) - lambda(j) ) * u(:,i)*u(:,j)';
                    end
                end
            end
            N = -N;
            rho = log(trace(temp));
            N = N + rho * eye(m);
            
            
            
            
            %% kelley's method (not with respect to V)
            n_init = 200;
            value = [];
            value_upper = [];
            m = size(N,1);
            V0 = eye(m); N0 = N; Y0 = 0;
            V = V0;
            N = N0;
            Y = Y0;
            grad_Vs = {};
            grad_Ns = {};
            grad_Ys = {};
            Delta = rho * 2;
            maxiter_subg = 1024;
            for iter_subg = 1:maxiter_subg
                N = (N+N')/2;
                V = (V+V')/2;
                Y = (Y+Y')/2;
                Vs{iter_subg} = V;
                Ns{iter_subg} = N;
                Ys{iter_subg} = Y;
                
                [s, lambdamax, umax] = compute_sigma_fdiv_lambdamax(H+Y,N,V,n_init);
                r = s / (1-s);
                value(iter_subg) =    -trace(B*N) + 2 * lambdamax;
                grad_Vs{iter_subg} =  - 2*(1-s)*  (r*log(r)-r+1) * umax*umax';
                grad_Ns{iter_subg} = - B + 2*(1-s)* umax*umax';
                grad_Ys{iter_subg} = 2*s*umax*umax';
                
                
                
                mloc = size(TPg,2);
                
                cvx_begin quiet
                %variable DeltaN(mloc,1)
                variable DeltaN(m,m)
                variable DeltaY(m,m)
                variable t(1,1);
                minimize ( t )
                subject to
                for i=1:iter_subg
                    % value(i) +  trace(grad_Ns{i}*(N0+reshape(TPg*DeltaN,[m m])-Ns{i})) <= t;
                    value(i) +  trace(grad_Ns{i}*(N0+ DeltaN -Ns{i}))+  trace(grad_Ys{i}*(Y0+ DeltaY -Ys{i})) <= t;
                end
                TPg' * DeltaY(:) == 0
                norm(DeltaN) + norm(DeltaY) <= Delta
                % norm(reshape(TPg*DeltaN,[m m])) <= Delta
                cvx_end
                
                % DeltaN = reshape(TPg*DeltaN,[m m]);
                if mod(iter_subg,m+2)==0
                    if  norm(DeltaN) + norm(DeltaY) >= Delta * .999, 'increasing Delta', Delta = Delta *2;
                    end
                end
                N = N0+DeltaN;
                Y = Y0+DeltaY;
                %N = N0 + DeltaN;
                value_upper(iter_subg) = t;
                if ( norm(DeltaN)+ norm(DeltaY)  < Delta * .999)
                    current_gaps(iter_subg) = min(value)-t;
                else
                    current_gaps(iter_subg) = Inf;
                end
                if (  norm(DeltaN) + norm(DeltaY) < Delta * .999) & ( ( min(value) - t ) < 1e-3 ), 'exit', break; end
            end
            
            [a,b] = min(value);
            QT_kelley_notV(irr) = a
            N = Ns{b};
            Y = Ys{b};
            
            all_gaps_noV{irr} = current_gaps;
            
            
            
            
            
            %% kelley's method (also with respect to V)
            n_init = 200;
            value = [];
            value_upper = [];
            m = size(N,1);
            V0 = eye(m); N0 = N; Y0 = Y;
            V = V0;
            N = N0;
            Y = Y0;
            grad_Vs = {};
            grad_Ns = {};
            grad_Ys = {};
            Delta = rho * 2;
            maxiter_subg = 1024;
            for iter_subg = 1:maxiter_subg
                N = (N+N')/2;
                V = (V+V')/2;
                Y = (Y+Y')/2;
                Vs{iter_subg} = V;
                Ns{iter_subg} = N;
                Ys{iter_subg} = Y;
                
                [s, lambdamax, umax] = compute_sigma_fdiv_lambdamax(H+Y,N,V,n_init);
                r = s / (1-s);
                value(iter_subg) =    -trace(B*N) + 2 * lambdamax;
                grad_Vs{iter_subg} =  - 2*(1-s)*  (r*log(r)-r+1) * umax*umax';
                grad_Ns{iter_subg} = - B + 2*(1-s)* umax*umax';
                grad_Ys{iter_subg} = 2*s*umax*umax';
                
                
                
                mloc = size(TPg,2);
                
                cvx_begin quiet
                %variable DeltaN(mloc,1)
                variable DeltaN(m,m)
                variable DeltaY(m,m)
                variable DeltaV(m,m)
                variable t(1,1);
                minimize ( t )
                subject to
                for i=1:iter_subg
                    % value(i) +  trace(grad_Ns{i}*(N0+reshape(TPg*DeltaN,[m m])-Ns{i})) <= t;
                    value(i) +  trace(grad_Ns{i}*(N0+ DeltaN -Ns{i}))+  trace(grad_Ys{i}*(Y0+ DeltaY -Ys{i})) <= t;
                end
                TPg' * DeltaY(:) == 0
                TPg' * DeltaV(:) == 0
                V0 + DeltaV == semidefinite(m)
                norm(DeltaN) + norm(DeltaY)  + norm(DeltaV)  <= Delta
                % norm(reshape(TPg*DeltaN,[m m])) <= Delta
                cvx_end
                
                % DeltaN = reshape(TPg*DeltaN,[m m]);
                if mod(iter_subg,m+2)==0
                    if  norm(DeltaN) + norm(DeltaY)  + norm(DeltaV)  >= Delta * .999, 'increasing Delta', Delta = Delta *2;
                    end
                end
                N = N0+DeltaN;
                Y = Y0+DeltaY;
                V = V0+DeltaV;
                %N = N0 + DeltaN;
                value_upper(iter_subg) = t;
                if ( norm(DeltaN)+ norm(DeltaY) + norm(DeltaV) < Delta * .999)
                    current_gaps(iter_subg) = min(value)-t;
                else
                    current_gaps(iter_subg) = Inf;
                end
                if (  norm(DeltaN) + norm(DeltaY) + norm(DeltaV) < Delta * .999) & ( ( min(value) - t ) < 1e-3 ), 'exit', break; end
            end
            
            
            QT_kelley_V(irr) = min(value)
            all_gaps_V{irr} = current_gaps;
            
        end
    end
    
    
    
    save(sprintf('data_logpartition_%s',model))
    
    try
        openfig('saved_figure.fig');
    catch
        'figure file not found'
    end
    
    %% plot
    plot(rrrs,QT_no_diag_recentered*0+logparval,'g','linewidth',2);
    hold on
    
    plot(rrrs,QT_kelley_V,'bx-','linewidth',2,'markersize',8);
    plot(rrrs,QT_kelley_notV,'bx:','linewidth',2,'markersize',8);
    plot(rrrs,QT_no_diag_recentered,'kx-','linewidth',2,'markersize',8);
    
    legend('log-partition-function','QT', 'QT - no V','QT - adapted from [4]','location','best');
    hold off
    set(gca,'fontsize',18)
    ylabel('log-partition functions');
    xlabel('maximal frequency (r)');
    axis([0 max(rrrs) 0 1])
    print('-depsc', sprintf('figure_logpartition_%s.eps',model));
end


